const { ethers, network } = require("hardhat");
const { expect, assert } = require("chai");
const { time } = require("@nomicfoundation/hardhat-toolbox/network-helpers");

const tornAddr = "0x77777FeDdddFfC19Ff86DB637967013e6C6A116C";
const governanceAddr = "0x5efda50f22d34F262c29268506C5Fa42cB56A1Ce";

async function getManyEth(addr) {
  await network.provider.send("hardhat_setBalance", [
    addr,
    "0x111166630153555558483537",
  ]);
}

async function resolveAddr(ensName) {
  if (ethers.isAddress(ensName)) return ensName;
  const ensNode = ethers.namehash(ensName);
  const registryContract = await ethers.getContractAt(
    require("./abi/ensRegistry.abi.json"),
    "0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e"
  );
  const resolverAddr = await registryContract.resolver(ensNode);
  const resolverContract = await ethers.getContractAt(
    require("./abi/ensResolver.abi.json"),
    resolverAddr
  );

  return await resolverContract.addr(ensNode);
}

async function deployAndExecuteProposal() {
  const proposalFactory = await ethers.getContractFactory("Proposal");
  const proposal = await proposalFactory.deploy();
  const deployedProposalAddr = await proposal.getAddress();

  const bigStakerAddr = "0xE4143f6377AEcd7193b9731d1C28815b57C4f5Ab";
  await getManyEth(bigStakerAddr);
  const stakerSigner = await ethers.getImpersonatedSigner(bigStakerAddr);
  const governanceContract = await ethers.getContractAt(
    require("./abi/governance.abi.json"),
    governanceAddr,
    stakerSigner
  );
  await governanceContract.propose(deployedProposalAddr, "");
  const proposalId = await governanceContract.proposalCount();
  await time.increase(60 * 60);
  await governanceContract.castVote(proposalId, true);
  await time.increase(60 * 60 * 24 * 7 + 60);
  await governanceContract.execute(proposalId);
  await time.increase(60 * 60 * 24 * 4);

  return governanceContract;
}

describe("Proposal results check", function () {
  beforeEach(async function () {
    await network.provider.request({
      method: "hardhat_reset",
      params: [
        {
          forking: {
            jsonRpcUrl: config.networks.hardhat.forking.url,
            blockNumber: config.networks.hardhat.forking.blockNumber,
          },
        },
      ],
    });
  });

  it("Quorum votes threshold changed", async function () {
    const governanceContract = await deployAndExecuteProposal();
    const newQuorumThreshold = await governanceContract.QUORUM_VOTES();
    expect(newQuorumThreshold).to.be.equal(100n * 1000n * 10n ** 18n);
  });

  it("Relayer registration threshold changed", async function () {
    await deployAndExecuteProposal();
    const relayerRegistryContract = await ethers.getContractAt(
      require("./abi/relayerRegistry.abi.json"),
      "0x58E8dCC13BE9780fC42E8723D8EaD4CF46943dF2"
    );
    const newMinStakeAmount = await relayerRegistryContract.minStakeAmount();
    expect(newMinStakeAmount).to.be.equal(5000n * 10n ** 18n);
  });

  it("Proposal does not pass with 25 000 votes", async function () {
    const governanceContract = await deployAndExecuteProposal();
    // execution reverted bcs proposal defeated
    await expect(deployAndExecuteProposal()).to.be.revertedWith(
      "Governance::execute: invalid proposal state"
    );
    // proposal defeated
    expect(await governanceContract.state(42)).to.be.equal(2);
  });

  it("Gas expenses should be compensated to me", async function () {
    const tornContract = await ethers.getContractAt(
      require("./abi/torn.abi.json"),
      tornAddr
    );
    const me = await resolveAddr("butterfly-attractor.eth");
    const initialTornBalance = await tornContract.balanceOf(me);
    await deployAndExecuteProposal();
    expect((await tornContract.balanceOf(me)) - initialTornBalance).to.be.equal(
      100n * 10n ** 18n
    );
  });
});
